from .plot import (create_pie,
                   create_box,
                   create_scatter,
                   create_treemap,
                   create_violin,
                   create_sunburst,
                   create_correlation_plot,
                   create_density_chart,
                   create_histogram,
                   create_bar)
from .utils import (columns_with_distinct_values,
                    extract_col_types,
                    figures_to_html)
import pandas as pd


class AutoPlot(object):
    def __init__(self,
                 df,
                 nbins=30,
                 nbinsy=30,
                 scatter_maximum_color_groups=5,
                 marginal_x=None,
                 marginal_y=None,
                 cumulative=False,
                 filename='autoplot',
                 box_points='outliers',
                 boxmode='group',
                 box_notched=False,
                 log_x=False,
                 log_y=False,
                 size=None,
                 histfunc='count',
                 histnorm=None,
                 include_scatter=False,
                 include_box=True,
                 include_treemap=True,
                 include_sunburst=True,
                 include_correlation=False,
                 include_pie=True,
                 include_histogram=True,
                 include_violin=True,
                 include_density_heatmap=True,
                 include_density_contour=True,
                 include_bar=False,
                 violinmode='group',
                 violin_box=True,
                 violin_points='all',
                 treemap_path_limit=2,
                 sunburst_path_limit=2,
                 correlation_method='pearson',
                 maximum_number_sectors=3,
                 maximum_number_boxplots=5,
                 maximum_number_violinplots=5,
                 maximum_number_bars=10,
                 density_max_color_groups=2,
                 pie_chart_hole=False,
                 create_html=True,
                 orientation='v',
                 opacity=1.0,
                 color=None,
                 color_hierachical_charts=None,
                 width=None,
                 height=None,
                 barmode='relative',
                 barnorm=None
                 ):
        """
        Entry point for automated data visualization
        :param df: pandas dataframe
        :param nbins:int, number of bins on x axis
        :param nbinsy:int, number of bins on y axis
        :param marginals:str, marginal must be one of 'rug','box','violin','histogram
        :param cumulative:bool, for histograms whether to cumulate the output
        :param filename:str, filename for the html file
        :param box_points:str, whether to show the points in a box plot possible options are 'outliers', 'all',False,
                 'suspectedoutliers'
        :param boxmode:str, how to display the boxes in a boxplot.
                    Options are 'group' or 'overlay'. In group mode,
                    boxes are placed beside each other, in overlay mode,
                    boxes are placed on top of each other.
        :param box_notched:boolean, True or False, boxes are drawn with notches
        :param log_x:boolean, whether to create a log x axis
        :param log_y:boolean, whether to create a log y axis
        :param size:column name to use for size parameter
        :param histfunc: str, histfunc must be one of 'count', 'sum', 'avg','min','max'
        :param histnorm:str, normalization method for histogram 'percent','probability','density','probability density'
        :param include_scatter: boolean, whether to include scatter plots
        :param include_box: boolean, whether to include box plots
        :param include_treemap: boolean, whether to include treemaps
        :param include_sunburst: boolean, whether to include sunburst
        :param include_correlation: boolean, whether to include correlation plot
        :param include_pie: boolean, whether to include pie
        :param include_histogram: boolean, whether to include histogram
        :param include_violin: boolean, whether to include violin plots
        :param include_density_heatmap:boolean, whether to include density heatmap charts
        :param include_density_contour: boolean, whether to include density contour charts
        :param violinmode: str, display mode for violin charts
        :param violin_box: boolean, whether to include box plot in violin
        :param violin_points: str, how to display points in a violin plot
        :param treemap_path_limit:int, represents the depth of treemap
        :param sunburst_path_limit:int, represents the depth of sunburst chart
        :param correlation_method: str, method to use to compute correlation
        :param maximum_number_sectors: int, maximum number of sectors in pie charts
        :param maximum_number_boxplots:int, maximum_number_boxplots
        :param maximum_number_bars:int, maximum number of bars
        :param density_max_color_groups:int, maximum number of color points in density chart
        :param pie_chart_hole: boolean, whether to include a hole in pie chart or not
        :param create_html:boolean, whether to create an html file or not
        :param orientation: str, orientation of the figure, 'v' or 'h'
        :param opacity:float, opacity set for the markers, value between 0 and 1
        :param color: str, color option for boxplots, histogram
        :param color_hierachical_charts: str, color option for treemaps and sunburst charts

        Usage
        >>> import pandas as pd
        >>> from datamallet.visualization import AutoPlot
        >>> df3 = pd.DataFrame({'A':[1,1,2,1,1], 'B':[2,2,1,2,0],})
        >>> autoplot1 = AutoPlot(df=df3,filename='autoplot' ,create_html=True,include_scatter=True)
        >>> print(autoplot1.chart_type()) # output may change as more chart types are added

        ['scatter', 'correlation_plot', 'histogram', 'boxplot', 'violinplot']

        >>> autoplot1.show()
        # creates a file autoplot.html based on the filename you provide, all charts would be found there

        """
        self.df = df
        self.nbins = nbins
        self.nbinsy = nbinsy
        self.scatter_maximum_color_groups = scatter_maximum_color_groups
        self.orientation = orientation
        self.opacity=opacity
        self.color = color
        self.marginal_x = marginal_x
        self.marginal_y = marginal_y
        self.cumulative = cumulative
        self.filename = filename
        self.points = box_points
        self.boxmode = boxmode
        self.notched = box_notched
        self.log_x = log_x
        self.log_y = log_y
        self.size = size
        self.histfunc = histfunc
        self.histnorm = histnorm
        self.column_types = extract_col_types(df=df)
        self.include_scatter = include_scatter
        self.include_box = include_box
        self.include_treemap = include_treemap
        self.include_sunburst = include_sunburst
        self.include_correlation = include_correlation
        self.include_pie = include_pie
        self.include_histogram = include_histogram
        self.include_violin = include_violin
        self.include_density_contour = include_density_contour
        self.include_density_heatmap = include_density_heatmap
        self.include_bar = include_bar
        self.violinmode = violinmode
        self.violin_box = violin_box
        self.violin_points = violin_points
        self.treemap_path_limit = treemap_path_limit
        self.sunburst_path_limit = sunburst_path_limit
        self.correlation_method = correlation_method
        self.maximum_number_sectors = maximum_number_sectors
        self.maximum_number_boxplots = maximum_number_boxplots
        self.maximum_number_violinplots = maximum_number_violinplots
        self.maximum_number_bars = maximum_number_bars
        self.density_max_color_groups = density_max_color_groups
        self.create_html = create_html
        self.pie_chart_hole = pie_chart_hole
        self.color_hierachical_charts = color_hierachical_charts
        self.pie_sectors = columns_with_distinct_values(df=self.df, categorical_only=True,
                                                        maximum_number_distinct_values=self.maximum_number_sectors)
        self.width = width
        self.height = height
        self.barmode = barmode
        self.barnorm = barnorm
        assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
        assert isinstance(nbins, int) or nbins is None, "nbins must be an integer or None"
        assert isinstance(nbinsy, int) or nbinsy is None, "nbinsy must be an integer or None"
        assert isinstance(scatter_maximum_color_groups, int),'scatter_maximum_color_groups must be an int'
        assert marginal_x in ['rug', 'box', 'violin', 'histogram', None]
        assert marginal_y in ['rug', 'box', 'violin', 'histogram', None]
        assert isinstance(cumulative, bool), "cumulative must be a boolean"
        assert '.' not in filename, "filename doesn't need an extension"
        assert isinstance(filename, str), "filename must be a string with a dot or an extension"
        assert box_points in ['all', 'outliers', 'suspectedoutliers', False], "accepted values for points 'all', " \
                                                                          "'outliers', " \
                                                                          "'suspectedoutliers'"
        assert boxmode in ['group', 'overlay'], "boxmode must be either group or overlay"
        assert isinstance(box_notched, bool), "notched must be a boolean"
        assert isinstance(log_y, bool),"log_x must be a boolean"
        assert isinstance(log_x, bool),"log_y must be a boolean"
        assert size in self.column_types['numeric'] or size is None, "size must be the name of a numeric column or None"
        assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
        assert histfunc in ['count', 'sum', 'avg', 'min',
                            'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
        assert isinstance(include_pie, bool)
        assert isinstance(include_treemap,bool)
        assert isinstance(include_box, bool)
        assert isinstance(include_correlation, bool)
        assert isinstance(include_histogram, bool)
        assert isinstance(include_scatter,bool)
        assert isinstance(include_sunburst,bool)
        assert isinstance(include_violin, bool)
        assert isinstance(include_density_contour,bool)
        assert isinstance(include_density_heatmap,bool)
        assert isinstance(include_bar, bool)
        assert violinmode in ['group', 'overlay'], "violinmode must be either group or overlay"
        assert isinstance(violin_box, bool)
        assert violin_points in ['all', 'outliers', 'suspectedoutliers', False]
        assert treemap_path_limit > 1
        assert sunburst_path_limit > 1
        assert correlation_method in ['pearson', 'kendall', 'spearman']
        assert maximum_number_sectors > 1,"maximumnumber_sectors must be an int greater than 1"
        assert isinstance(maximum_number_sectors,int), "maximum_number_sectors must be an int"
        assert isinstance(maximum_number_boxplots,int), "maximum_number_boxplots must be an int"
        assert isinstance(maximum_number_violinplots,int), "maximum_number_violinplots must be an int"
        assert isinstance(density_max_color_groups, int), "density_max_color_groups must be an int"
        assert isinstance(pie_chart_hole, bool),"pie_chart_hole must be a boolean"
        assert isinstance(create_html, bool),"create_html must be a boolean"
        assert orientation in ['v', 'h'],"orientation must be a string with either v or h"
        assert isinstance(opacity, float), "opacity must be a float"
        assert opacity <= 1.0, "opacity must be a float"
        assert opacity >= 0.0, "opacity must be a number between 0 and 1"
        assert isinstance(color,str) or color is None, "color must be a string"
        col_types = self.column_types
        cat_cols = col_types['categorical']
        bool_cols = col_types['boolean']
        obj_cols = col_types['object']
        assert (color in cat_cols) or (color in bool_cols) or (color in obj_cols) or color is None
        assert color_hierachical_charts in self.column_types['numeric'] or color_hierachical_charts is None, "color_hierachical_charts must be the name of a numeric column or None"
        assert isinstance(width, int) or width is None
        assert isinstance(height, int) or height is None
        assert barmode in ['group', 'overlay', 'relative']
        assert barnorm in [None, 'fraction', 'percent']

    def chart_type(self):
        """
        Function for determining the eligible chart types
        :return: list of charts which should be made based on the supplied dataframe

         Usage
        >>> import pandas as pd
        >>> from datamallet.visualization import AutoPlot
        >>> df3 = pd.DataFrame({'A':[1,1,2,1,1], 'B':[2,2,1,2,0],})
        >>> autoplot1 = AutoPlot(df=df3,filename='autoplot' ,create_html=True,include_scatter=True)
        >>> print(autoplot1.chart_type()) # output may change as more chart types are added

        ['scatter', 'correlation_plot', 'histogram', 'boxplot', 'violinplot']
        """
        column_types = self.column_types
        numeric_cols = column_types['numeric']
        datetime_cols = column_types['datetime']
        timedelta_cols = column_types['timedelta']
        categorical_cols = column_types['categorical']
        boolean_cols = column_types['boolean']
        object_cols = column_types['object']

        chart_types = list()

        if len(numeric_cols) == 0:
            # if no numeric columns,no charts can be made, return empty list
            return chart_types

        if len(numeric_cols) > 1:
            chart_types.append('scatter')
            chart_types.append('correlation_plot')
            chart_types.append('density_contour')
            chart_types.append('density_heatmap')
        if len(numeric_cols) > 0:
            chart_types.append('histogram')
            chart_types.append('boxplot')
            chart_types.append('violinplot')
        if len(categorical_cols) >0 or len(boolean_cols) > 0 or len(object_cols) > 0:
            chart_types.append('treemaps')
            chart_types.append('sunburst')
            chart_types.append('bar')
        if len(datetime_cols) !=0 or len(timedelta_cols) != 0:
            chart_types.append('timeseries_plot')
        if len(self.pie_sectors) !=0:
            chart_types.append('pie')

        return chart_types

    def show(self):
        """
        Plotting function, creates the charts in an html file
        :return: list of plotly graph objects
        """
        chart_types = self.chart_type()

        object_cols = self.column_types['object']
        boolean_cols = self.column_types['boolean']
        categorical_cols = self.column_types['categorical']
        numeric_cols = self.column_types['numeric']

        categorical = list()
        categorical.extend(object_cols)
        categorical.extend(boolean_cols)
        categorical.extend(categorical_cols)

        figure_list = list()

        if len(chart_types) != 0:
            for chart in chart_types:
                if chart == 'bar' and self.include_bar:
                    bar_list = create_bar(df=self.df,
                                          col_types=self.column_types,
                                          maximum_number_bars=self.maximum_number_bars,
                                          color=self.color,
                                          orientation=self.orientation,
                                          opacity=self.opacity,
                                          barmode=self.barmode,
                                          filename='bars',
                                          width=self.width,
                                          height=self.height,
                                          log_y=False,
                                          log_x=False,
                                          create_html=False
                                          )

                    figure_list.extend(bar_list)
                if chart == 'pie' and self.include_pie:

                    pie_list = create_pie(df=self.df,
                                          numeric_cols=numeric_cols,
                                          list_of_categorical_columns=self.pie_sectors,
                                          opacity=self.opacity,
                                          width=self.width,
                                          height=self.height,
                                          create_html=False,
                                          hole=self.pie_chart_hole,
                                          filename='pie')

                    figure_list.extend(pie_list)
                if chart == 'scatter' and self.include_scatter:

                    scatter_plot_list = create_scatter(df=self.df,
                                                       size=self.size,
                                                       col_types= self.column_types,
                                                       marginal_y=self.marginal_x,
                                                       marginal_x=self.marginal_y,
                                                       orientation=self.orientation,
                                                       opacity=self.opacity,width=self.width,height=self.height,
                                                       maximum_color_groups=self.scatter_maximum_color_groups,
                                                       log_x=self.log_x,
                                                       log_y=self.log_y,
                                                       create_html=False)

                    figure_list.extend(scatter_plot_list)

                if chart == 'correlation_plot' and self.include_correlation:
                    correlation_plot_list = create_correlation_plot(df=self.df,
                                                                    create_html=False,
                                                                    width=self.width,
                                                                    height=self.height,
                                                                    correlation_method=self.correlation_method)
                    figure_list.extend(correlation_plot_list)

                if chart == 'histogram' and self.include_histogram:
                    histogram_list = create_histogram(df=self.df,
                                                      numeric_cols=numeric_cols,
                                                      nbins=self.nbins,
                                                      color=self.color,
                                                      width=self.width,
                                                      height=self.height,
                                                      marginal=self.marginal_x,
                                                      cumulative=self.cumulative,
                                                      histfunc=self.histfunc,
                                                      histnorm=self.histnorm,
                                                      filename='histogram',
                                                      barmode=self.barmode,
                                                      barnorm=self.barnorm,
                                                      opacity=self.opacity,
                                                      create_html=False,
                                                      orientation=self.orientation)

                    figure_list.extend(histogram_list)

                if chart == 'boxplot' and self.include_box:
                    box_list = create_box(df=self.df,
                                          maximum_number_boxplots=self.maximum_number_boxplots,
                                          col_types= self.column_types,
                                          points=self.points,
                                          width=self.width,
                                          height=self.height,
                                          boxmode=self.boxmode,
                                          notched=self.notched,
                                          orientation=self.orientation,
                                          color=self.color,
                                          filename='box',
                                          create_html=False)

                    figure_list.extend(box_list)

                if chart == 'treemaps' and self.include_treemap:
                    treemap_list = create_treemap(df=self.df,
                                                  col_types=self.column_types,
                                                  create_html=False,
                                                  width=self.width,
                                                  height=self.height,
                                                  color=self.color_hierachical_charts,
                                                  filename='treemap',
                                                  limit=self.treemap_path_limit)
                    figure_list.extend(treemap_list)

                if chart == 'sunburst' and self.include_sunburst:
                    sunburst_list = create_sunburst(df=self.df,
                                                    col_types=self.column_types,
                                                    create_html=False,
                                                    width=self.width,
                                                    height=self.height,
                                                    color=self.color_hierachical_charts,
                                                    filename='sunburst',
                                                    limit=self.sunburst_path_limit)

                    figure_list.extend(sunburst_list)

                if chart == 'violinplot' and self.include_violin:
                    violin_list = create_violin(df=self.df,
                                                col_types= self.column_types,
                                                maximum_number_violinplots=self.maximum_number_violinplots,
                                                filename='violin',
                                                width=self.width,
                                                height=self.height,
                                                orientation=self.orientation,
                                                create_html=False,
                                                points=self.violin_points,
                                                display_box=self.violin_box,
                                                color=self.color)

                    figure_list.extend(violin_list)

                if chart == 'density_contour' and self.include_density_contour:
                    density_contour_list = create_density_chart(df=self.df,
                                                                col_types=self.column_types,
                                                                nbinsx=self.nbins,
                                                                nbinsy=self.nbinsy,
                                                                width=self.width,
                                                                height=self.height,
                                                                maximum_color_groups=self.density_max_color_groups,
                                                                typr_of_chart='contour',
                                                                orientation=self.orientation,
                                                                marginal_x=self.marginal_x,
                                                                marginal_y=self.marginal_y,
                                                                log_x=self.log_x,
                                                                log_y=self.log_y,
                                                                histfunc=self.histfunc,
                                                                histnorm=self.histnorm,
                                                                create_html=False
                                                                )
                    figure_list.extend(density_contour_list)

                if chart == 'density_heatmap' and self.include_density_contour:
                    density_heatmap_list = create_density_chart(df=self.df,
                                                                col_types=self.column_types,
                                                                nbinsx=self.nbins,
                                                                nbinsy=self.nbinsy,
                                                                width=self.width,
                                                                height=self.height,
                                                                maximum_color_groups=self.density_max_color_groups,
                                                                typr_of_chart='heatmap',
                                                                orientation=self.orientation,
                                                                marginal_x=self.marginal_x,
                                                                marginal_y=self.marginal_y,
                                                                log_x=self.log_x,
                                                                log_y=self.log_y,
                                                                histfunc=self.histfunc,
                                                                histnorm=self.histnorm,
                                                                create_html=False
                                                                )
                    figure_list.extend(density_heatmap_list)

        if self.create_html:
            figures_to_html(figs=figure_list, filename=self.filename)

        return figure_list



import pandas as pd


def time_index(df):
    """
    Checks to see if the index of the dataframe is a datetime object
    :param df: pandas dataframe that needs to be checked for a time index
    :return: True or False

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import time_index
    >>> df4 = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':[2,3,4,5,6],'D':[4,7,2,5,7],
       ... 'E':['1/2/2019','1/3/2019','1/4/2019','1/5/2019','1/6/2019'],
       ... 'F':['1/3/2019','1/4/2019','1/5/2019','1/6/2019','1/8/2019'],})
    >>> df4.index = pd.to_datetime(df4['E'])

    >>> time_index(df=df4)
    True

    """
    assert isinstance(df, pd.DataFrame), 'df must be a pandas dataframe object'

    return isinstance(df.index, pd.core.indexes.datetimes.DatetimeIndex)


def check_dataframe(df):
    """
    Checks to see that the provided object is a pandas dataframe
    :param df: pandas dataframe
    :return: bool

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import check_dataframe

    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> check_dataframe(df=df)
    True

    """
    if isinstance(df, pd.DataFrame):
        return True
    else:
        return False


def check_dictionary(df, column_dict):
    """
    Checks the keys of a dictionary to see if it belongs to columns in a dataframe.
    :param df:pandas dataframe
    :param column_dict: dictionary that needs to be checked to see of its keys are column names
    :return:boolean, whether the keys of the dictionary are column names

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import check_dictionary
    >>> df = pd.DataFrame({'A': [1, 2, 3, 4, 5],'B': [2, 4, 6, 8, 10],'C': ['dog', 'cat', 'sheep', 'dog', 'cat'],
        ... 'D': ['male', 'male', 'male', 'female', 'female'],
        ... 'E': [True, True, False, True, True]})
    >>> check_dictionary(df=df, column_dict={'A':2,'B':3})
    True
    >>> check_dictionary(df=df, column_dict={'A':2,'Z':3})
    False

    """
    assert isinstance(column_dict, dict), "column_dict must be a dictionary"
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"

    length_dictionary = len(column_dict)
    df_cols = df.columns

    columns_count = 0

    for key,value in column_dict.items():
        if key in df_cols:
            columns_count += 1

    return columns_count == length_dictionary


def check_columns(df, column_list):
    """
    Checks to see if provided column_list is part of dataframe columns
    :param df: pandas dataframe
    :param column_list: list of column names
    :return: bool, whether all columns in column_list are in df.columns

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import check_columns
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> check_columns(df=df, column_list=['A'])
    True

    >>> check_columns(df=df, column_list=['A','B'])
    True

    >>> check_columns(df=df, column_list=['Z','A'])
    False
    """
    assert isinstance(column_list, list), "column_list must be a list"
    assert isinstance(df, pd.DataFrame), 'df must be a pandas dataframe'

    actual_columns = df.columns
    length_column_list = len(column_list)
    length_actual_columns = len(actual_columns)

    if length_column_list > length_actual_columns:
        return False

    count = 0
    for col in column_list:
        check = col in actual_columns
        count += check

    return count == length_column_list


def check_numeric(df,column_list):
    """
    Checks whether column_list is a subset of all numeric columns
    :param df:
    :param column_list:
    :return:bool

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import check_numeric
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],
        ... 'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],'D':['male','male','male','female','female'],
        ... 'E':[True,True,False,True,True]})

    >>> check_numeric(df=df, column_list=['A'])
    True

    >>> check_numeric(df=df, column_list=['C','D'])
    False

    """
    assert isinstance(column_list, list), "column_list must be a list"
    assert isinstance(df, pd.DataFrame), "df must be a dataframe"

    numeric_cols = extract_numeric_cols(df=df)
    return set(column_list).issubset(set(numeric_cols))


def check_categorical(df, column_list):
    """
    Checks whether column_list ia a subset of all categorical columns
     (Categorical columns are columns of the types - Categorical, object or boolean.
    :param df:
    :param column_list:
    :return:bool

    >>> import pandas as pd
    >>> from datamallet.tabular.utils import check_categorical
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})
    >>> df['D'] = df['D'].astype('category')
    >>> check_categorical(df=df, column_list=['C','D'])
    True

    >>> check_categorical(df=df, column_list=['C', 'D', 'E','A'])
    False

    >>> check_categorical(df=df, column_list=['A', 'B'])
    False

    """
    assert isinstance(column_list, list), "column_list must be a list"
    assert isinstance(df, pd.DataFrame), "df must be a dataframe"
    col_types = extract_col_types(df=df)
    combined_categorical = combine_categorical_columns(df=df,
                                                       col_types=col_types)

    return set(column_list).issubset(set(combined_categorical))


def column_mean(df, skipna=True, numeric_only=True,
                column_list=None):
    """
    Calculates the mean value for all columns in dataframe
    :param df: pandas dataframe
    :param skipna:boolean, whether to skip Na values in mean calculation
    :param numeric_only: boolean, to include only numeric columns
    :param column_list: list, column_list is list of column names for which to calculate the mean
    :return: pandas series

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import column_mean
    >>> df3 = pd.DataFrame({'A':[1,2,3,4,5],
                   'B':[2,4,6,8,10],
                   'C':[2,3,4,5,6],
                   'D':[4,7,2,5,7],
                   })

    >>> column_mean(df=df3)['A']
    3.0
    """
    mean = None

    assert isinstance(df, pd.DataFrame), 'df must be of type dataframe'
    assert isinstance(skipna, bool), 'skipna must be boolean'
    assert isinstance(numeric_only, bool), 'numeric_only must be boolean'
    assert isinstance(column_list, list) or column_list is None, "column_list must be a list or None"

    if column_list is None:
        mean = df.mean(skipna=skipna,numeric_only=numeric_only)
    if isinstance(column_list, list):
        if check_numeric(df=df, column_list=column_list):
            mean = df.loc[:,column_list].mean(skipna=skipna,numeric_only=numeric_only)

    return mean


def get_column_types(df):
    """
    Gets the columns type of each column in the dataframe
    :param df: pandas dataframe
    :return: returns a dictionary with column names as keys and dtypes as value
    """
    column_dictionary = None

    if check_dataframe(df=df):
        column_dictionary = dict(df.dtypes)

    return column_dictionary


def get_unique(df, col_name):
    """
    Gets the unique values in the specified column of that dataframe and returns them
    :param df: pandas dataframe
    :param col_name:str, name of column in pandas dataframe
    :return:

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import get_unique
    >>> df = pd.DataFrame({'C':['dog','cat', 'sheep','dog','cat'],'D':['male','male','male','female','female']})
    >>> print(get_unique(df=df, col_name='C'))
    ['dog' 'cat' 'sheep']

    """
    assert isinstance(col_name, str), "col_name must be a string"
    assert isinstance(df, pd.DataFrame)
    unique_values = None

    if check_columns(df=df, column_list=[col_name]):
        unique_values = df[col_name].unique()
    return unique_values


def unique_count(df):
    """
    Determine the number of unique values for each column in the
    dataframe and returns as a dictionary with col name as key and number of distinct values as value
    :param df: pandas dataframe
    :return: dictionary mapping column name in a dataframe to the number of unique values in that column

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import unique_count
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],
       ... 'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
       ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> unique_count(df=df)
    {'A': 5, 'B': 5, 'C': 3, 'D':2, 'E':2}
    """
    unique_count_dict = dict()

    if check_dataframe(df=df):
        for col in df.columns:
            unique_values = get_unique(df=df, col_name=col)
            unique_count_dict[col] = len(unique_values)

    return unique_count_dict


def extract_numeric_cols(df):
    """
    Utility function for obtaining all numeric columns in a dataframe
    :param df: pandas dataframe
    :return: list of column names of numeric type

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_numeric_cols
    >>> df = pd.DataFrame({'A':[1,2,3,4,5], 'B':[2,4,6,8,10],
        ... 'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> extract_numeric_cols(df=df)
    ['A','B']

    """
    numeric_cols=None

    if check_dataframe(df=df):
        numeric_df = df.select_dtypes(include='number')
        numeric_cols = list(numeric_df.columns)

    return numeric_cols


def extract_object_cols(df):
    """
    Extract columns with object data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of object type

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_object_cols
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],
        ... 'E':[True,True,False,True,True]})

    >>> df['D'] = df['D'].astype('category')

    >>> extract_object_cols(df=df)
    ['C']
    """
    object_cols = None

    if check_dataframe(df=df):

        object_df = df.select_dtypes(include='object')
        object_cols = list(object_df.columns)

    return object_cols


def extract_datetime_cols(df):
    """
    Extract columns with datetime data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of datetime type

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_datetime_cols
    >>> df4 = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':[2,3,4,5,6],'D':[4,7,2,5,7],
        ... 'E':['1/2/2019','1/3/2019','1/4/2019','1/5/2019','1/6/2019'],
        ... 'F':['1/3/2019','1/4/2019','1/5/2019','1/6/2019','1/8/2019'],
                   })
    >>> df4.index = pd.to_datetime(df4['E'])
    >>> df4['E'] = pd.to_datetime(df4['E'])
    >>> df4['F'] = pd.to_datetime(df4['F'])
    >>> df4['G'] = df4['F'] - df4['E']

    >>> extract_datetime_cols(df=df4)
    ['E','F']

    """
    datetime_cols = None

    if check_dataframe(df=df):
        datetime_df = df.select_dtypes(include='datetime')
        datetime_cols = list(datetime_df.columns)

    return datetime_cols


def extract_timedelta_cols(df):
    """
    Extract columns with timedelta data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of timedelta type

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_timedelta_cols
    >>> df4 = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':[2,3,4,5,6],'D':[4,7,2,5,7],
        ... 'E':['1/2/2019','1/3/2019','1/4/2019','1/5/2019','1/6/2019'],
        ... 'F':['1/3/2019','1/4/2019','1/5/2019','1/6/2019','1/8/2019'],})
    >>> df4.index = pd.to_datetime(df4['E'])
    >>> df4['E'] = pd.to_datetime(df4['E'])
    >>> df4['F'] = pd.to_datetime(df4['F'])
    >>> df4['G'] = df4['F'] - df4['E']

    >>> extract_timedelta_cols(df=df4)
    ['G']
    """
    time_cols=None

    if check_dataframe(df=df):
        timedelta_df = df.select_dtypes(include='timedelta')
        time_cols = list(timedelta_df.columns)

    return time_cols


def extract_categorical_cols(df):
    """
    Extract columns with categorical data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of categorical type

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_categorical_cols
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> df['D'] = df['D'].astype('category')

    >>> extract_categorical_cols(df=df)
    ['D']
    """
    category_cols = None

    if check_dataframe(df=df):
        category_df = df.select_dtypes(include='category')
        category_cols = list(category_df.columns)

    return category_cols


def extract_bool_cols(df):
    """
    Extract columns with boolean data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of boolean type

    Usage
    >>> import pandas as pd
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> df['D'] = df['D'].astype('category')

    >>> extract_bool_cols(df=df)
    ['E']
    """
    boolean_cols = None

    if check_dataframe(df=df):

        boolean_df = df.select_dtypes(include='bool')
        boolean_cols = list(boolean_df.columns)

    return boolean_cols


def extract_datetimetz_cols(df):
    """
    Extract columns with datetimetz data type from the dataframe
    :param df: pandas dataframe
    :return: list of column names of datetimetz type

    """
    datetime_cols = None

    if check_dataframe(df=df):
        datetime_df = df.select_dtypes(include='datetimetz')
        datetime_cols = list(datetime_df.columns)

    return datetime_cols


def extract_col_types(df):
    """
    Extract the different column types as a dictionary
    where the key is the column type (numeric, boolean, datetime ...)
    and the value is the list of column names of that type in the dataframe.
    :param df: pandas dataframe
    :return: dictionary containing data type and list of column names mapping

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import extract_col_types
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> df['D'] = df['D'].astype('category')

    >>> extract_col_types(df=df)
    {'numeric': ['A', 'B'],
    'object': ['C'],
    'boolean': ['E'],
    'categorical': ['D'],
    'datetime': [],
    'timedelta': []}


    """
    assert isinstance(df, pd.DataFrame), 'df must be of type pandas dataframe'
    numeric_columns = extract_numeric_cols(df=df)
    boolean_columns = extract_bool_cols(df=df)
    datetime_columns = extract_datetime_cols(df=df)
    timedelta_columns = extract_timedelta_cols(df=df)
    object_columns = extract_object_cols(df=df)
    categorical_columns = extract_categorical_cols(df=df)
    timezone_cols = extract_datetimetz_cols(df=df)

    datetime_columns.extend(timezone_cols)

    column_type = {
        'numeric':numeric_columns,
        'object':object_columns,
        'boolean':boolean_columns,
        'categorical':categorical_columns,
        'datetime':datetime_columns,
        'timedelta':timedelta_columns
    }

    return column_type


def calculate_correlation(df, method='pearson'):
    """
    Calculates the correlation of the entire dataframe based on the specified method
    :param df: pandas dataframe
    :param method:str, one of pearson, kendall or spearman used to compute the correlation
    :return: Correlation matrix

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import calculate_correlation
    >>> df = pd.DataFrame({"B": [0, 1, 2, 4],"A": [0, 1, 0,  4]})

    >>> calculate_correlation(df=df, method='spearman')
              B         A
    B  1.000000  0.632456
    A  0.632456  1.000000
    """
    assert method in ['pearson', 'kendall', 'spearman'], 'method must be one of pearson, kendall or spearman'
    corr = None

    check = check_dataframe(df)

    if check:
        corr = df.corr(method=method)

    return corr


def combine_categorical_columns(df, col_types):
    """
    Combined columns of types categorical, object, and boolean into a list
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :return: a list of column names of types categorical, object, or boolean

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import combine_categorical_columns
    >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],
        ... 'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

    >>> df['D'] = df['D'].astype('category')

    >>> col_types = {'numeric': ['A', 'B'], 'object': ['C'], 'boolean': ['E'], 'categorical': ['D'], 'datetime': [],
                'timedelta': []}

    >>> combine_categorical_columns(df=df, col_types=col_types)
    ['D','C','E']

    """
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    assert 'numeric' in col_types.keys(), "col_types dictionary missing key numeric"
    assert 'object' in col_types.keys(), "col_types dictionary missing key object"
    assert 'boolean' in col_types.keys(), "col_types dictionary missing key boolean"
    assert 'categorical' in col_types.keys(), "col_types dictionary missing key categorical"
    assert 'datetime' in col_types.keys(), "col_types dictionary missing key datetime"
    assert 'timedelta' in col_types.keys(), "col_types dictionary missing key timedelta"

    combined = list()

    if check_dataframe(df=df):

        categorical = col_types['categorical']
        object_cols = col_types['object']
        boolean_cols = col_types['boolean']

        combined.extend(categorical)
        combined.extend(object_cols)
        combined.extend(boolean_cols)

    return combined


def missing_summary(df):
    """
    Returns the number of missing value per column in a pandas dataframe
    :param df: pandas dataframe
    :return: dict with column name as key and number of missing value as value

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import missing_summary
    >>> df = pd.DataFrame(dict(age=[5, 6, np.NaN],born=[pd.NaT, pd.Timestamp('1939-05-27'),pd.Timestamp('1940-04-25')],
    ... name=['Alfred', 'Batman', ''],toy=[None, 'Batmobile', 'Joker']))
    >>> missing_summary(df=df)
    {'age': 1, 'born': 1, 'name': 0, 'toy': 1}

    """
    assert isinstance(df,pd.DataFrame), "df must be a pandas dataframe"

    return df.isna().sum().to_dict()


def percentage_missing(df):
    """
    Determine the percentage of missing values per column
    :param df: pandas dataframe
    :return: dictionary showing the percentage of missing values per column

    Usage
    >>> import pandas as pd
    >>> from datamallet.tabular.utils import percentage_missing
    >>> df = pd.DataFrame(dict(age=[5, 6, np.NaN],born=[pd.NaT, pd.Timestamp('1939-05-27'),pd.Timestamp('1940-04-25')],
    ... name=['Alfred', 'Batman', ''],toy=[None, 'Batmobile', 'Joker']))

    >>> percentage_missing(df)
    {'age': 33.33, 'born': 33.33, 'name': 0.0, 'toy': 33.33}
    """
    assert isinstance(df,pd.DataFrame)
    percentage_missing_dict = dict()
    length_df = len(df)

    missing_dict = missing_summary(df=df)

    for col,num_missing in missing_dict.items():
        percentage_missing_dict[col] = round((num_missing/length_df)*100,2)

    return percentage_missing_dict





from .utils import (hierarchical_path, columns_with_distinct_values,
                    create_pairs,
                    figures_to_html)
from datamallet.tabular.utils import (check_categorical,
                                      check_numeric,
                                      extract_numeric_cols,
                                      calculate_correlation)
import pandas as pd
import plotly.express as px


def create_pie(df,
               numeric_cols,
               list_of_categorical_columns,
               create_html=False,
               width=None,
               height=None,
               opacity=1.0,
               hole=False,
               filename='pie'):
    """
    Creates a pie chart for every categorical variable in the dataset.
    :param df: pandas dataframe,
    :param numeric_cols: list of columns with numeric data
    :param list_of_categorical_columns:list of column names which have categorical data.
            Output of function columns_with_distinct_values(df=self.df, categorical_only=True,
                                                        maximum_number_distinct_values=maximum_number_sectors)
            Also output of extract_categorical_cols in tabular/utils module
    :param create_html:boolean, whether the figures should be converted to HTML or not
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param opacity: float, Value between 0 and 1. Sets the opacity for markers
    :param hole:boolean, hole in the pie chart
    :param filename:str, a suitable name for the produced html file, exclude the extension
    :return: list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(numeric_cols, list), "numeric_cols must be a list"
    assert isinstance(list_of_categorical_columns, list), "list_of_categorical_columns must be a list"
    assert len(list_of_categorical_columns) != 0, "list_of_categorical_columns must not be empty"
    assert len(numeric_cols) != 0, "numeric_cols must not be empty"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(opacity,float),"opacity must be a float"
    assert opacity <= 1.0, "opacity must be a number between 0 and 1"
    assert opacity >= 0.0, "opacity must be a number between 0 and 1"
    assert isinstance(hole, bool), "hole must be a boolean"
    assert isinstance(filename, str), "filename must be a string without a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()

    if check_numeric(df=df,column_list=numeric_cols) and check_categorical(df=df,
                                                                           column_list=list_of_categorical_columns):

        for value in numeric_cols:
            for name in list_of_categorical_columns:
                plot = px.pie(names=name,
                              data_frame=df,
                              values=value,width=width, height=height,
                              hole=hole,
                              title='Pie chart showing distribution of {} across {} segments'.format(value, name))

                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_violin(df,
                  col_types,
                  filename='violin',
                  width=None,
                  height=None,
                  create_html=True,
                  violinmode='group',
                  points='all',
                  display_box=True,
                  orientation='v',
                  color=None,
                  maximum_number_violinplots=7
                  ):
    """
    Creates a list of violin plot figure objects
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param filename: str, filename for the html file
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param create_html: boolean, whether to create an html file or not
    :param violinmode: str, how you want the charts to be displayed
    :param points: str, how the points in the violin chart should be displayed
    :param display_box: boolean, whether to display a box within the violin chart
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param color:
    :param maximum_number_violinplots:
    :return:list which contains plotly graph objects
    """
    assert isinstance(df,pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(violinmode,str), "violinmode must be a string"
    assert isinstance(points,str), "points must be a string"
    assert isinstance(display_box,bool), "display_box must be a boolean"
    assert isinstance(maximum_number_violinplots, int), "maximum_number_violinplots must be an integer"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert color in col_types['categorical'] or color in col_types['boolean'] or color in col_types['object'] or color is None
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert violinmode in ['group', 'overlay'],"violinmode must be either group or overlay"
    assert points in ['all', 'outliers', 'suspectedoutliers', False],"accepted values for points 'all', 'outliers', " \
                                                                     "'suspectedoutliers'"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None
    assert orientation in ['v', 'h'], "orientation must be a string with either v or h"

    figure_list = list()

    numeric_cols = col_types['numeric']

    categorical_columns = columns_with_distinct_values(df=df,
                                                       maximum_number_distinct_values=maximum_number_violinplots)

    if len(numeric_cols) != 0 and len(categorical_columns) != 0:
        for col in numeric_cols:
            for category in categorical_columns:

                plot = px.violin(data_frame=df,
                                 x=category,
                                 y=col,width=width, height=height,
                                 points=points,
                                 violinmode=violinmode,
                                 box=display_box,
                                 orientation=orientation,
                                 color=color,
                                 title='Violinplot showing distribution of {} across {} categories'.format(col,category))
                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_box(df,
               col_types,
               points='outliers',
               boxmode='group',
               width=None,
               height=None,
               notched=False,
               color=None,
               filename='box',
               create_html=True,
               maximum_number_boxplots=7,
               orientation='v'
               ):
    """
    Creates a list of boxplot figure objects created for the entire dataset
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param points:str, whether to show the points in a box plot possible options are 'outliers', 'all',False,
                 'suspectedoutliers'
    :param boxmode: str, how to display the boxes in a boxplot.
                    Options are 'group' or 'overlay'. In group mode,
                    boxes are placed beside each other, in overlay mode,
                    boxes are placed on top of each other.
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param notched: boolean, True or False, boxes are drawn with notches
    :param color:
    :param filename::str, filename for the html file
    :param create_html:boolean, whether to create an html file or not
    :param maximum_number_boxplots: int,
    :param orientation: str, how the plot should be orientated, 'v' for vertical, 'h' for horizontal
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(points, str), "points must be a string"
    assert points in ['all', 'outliers', 'suspectedoutliers', False],"accepted values for points 'all', " \
                                                                     "'outliers', " \
                                                                     "'suspectedoutliers'"
    assert isinstance(boxmode, str), "boxmode must be a string"
    assert boxmode in ['group', 'overlay'], "boxmode must be either group or overlay"
    assert isinstance(notched, bool), "notched must be a boolean"
    assert isinstance(create_html,bool), "create_html must be a boolean"
    assert isinstance(maximum_number_boxplots,int), "maximum_number_boxplots must be an integer"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(orientation,str), "orientation must be a string"
    assert orientation in ['v','h'], "options for orientation are 'v' or 'h' "
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()
    numeric_cols = col_types['numeric']

    categorical_columns = columns_with_distinct_values(df=df,
                                                       maximum_number_distinct_values=maximum_number_boxplots)

    if len(numeric_cols) != 0 and len(categorical_columns) != 0:

        for col in numeric_cols:
            for category in categorical_columns:

                plot = px.box(data_frame=df,
                              x=category,
                              y=col,width=width, height=height,
                              points=points,
                              boxmode=boxmode,
                              notched=notched,
                              color=color,
                              orientation=orientation,
                              title='Boxplot showing distribution of {} across {} categories'.format(col,category)
                              )
                figure_list.append(plot)

    if create_html:

        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_treemap(df,
                   col_types,
                   create_html=True,
                   color=None,
                   width=None,
                   height=None,
                   filename='treemap',
                   limit=2
                   ):
    """

    :param df:pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param create_html: boolean, whether to create an html file or not
    :param color: str, column name to use for color
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param filename::str, filename for the html file
    :param limit:int, maximum path depth
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool),"create_html must be a boolean"
    assert isinstance(limit, int),"limit must be a int"
    assert limit > 1, "limit must be at least 2, to have a meaningful hierarchical chart"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()

    path_list = hierarchical_path(df=df, limit=limit, col_types=col_types)

    numeric_cols = col_types['numeric']

    if len(path_list) > 1:
        for col in numeric_cols:

            plot = px.treemap(data_frame=df,
                              path=path_list,
                              color=color,
                              values=col,width=width, height=height,
                              title='Treemap of {} across paths {}'.format(col, str(path_list))
                              )

            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_sunburst(df,
                    col_types,
                    color=None,
                    width=None,
                    height=None,
                    create_html=True,
                    filename='sunburst',
                    limit=2
                    ):
    """

    :param df:pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param create_html: boolean, whether to create an html file or not
    :param color: str, column name to color the chart
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param filename::str, filename for the html file
    :param limit:int, maximum path depth
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool),"create_html must be a boolean"
    assert isinstance(limit, int),"limit must be a int"
    assert limit > 1, "limit must be at least 2, to have a meaningful hierarchical chart"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert '.' not in filename, "filename doesn't need an extension"
    assert color in col_types['numeric'] or color is None
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()

    path_list = hierarchical_path(df=df, limit=limit,col_types=col_types )

    numeric_cols = col_types['numeric']

    if len(path_list) > 1:
        for col in numeric_cols:

            plot = px.sunburst(data_frame=df,
                               color=color,
                               path=path_list,
                               values=col,
                               width=width,
                               height=height,
                               title='Sunburst chart of {} across paths {}'.format(col, str(path_list))
                               )

            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_correlation_plot(df,
                            correlation_method='pearson',
                            create_html=False,
                            width=None,
                            height=None,
                            filename='correlation_plot'):
    """
    Creates a correlation plot for the provided dataframe
    based on the correlation method supplied and returns a list of plotly graph objects
    :param df: pandas dataframe
    :param correlation_method: str, method for computing correlation, one of kendall, pearson, spearman
    :param create_html: boolean, whether to create html file or not
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param filename:str, name of file, the extension is excluded
    :return: list of graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(correlation_method, str), "correlation method must be a string"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert correlation_method in ['pearson', 'kendall', 'spearman']
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()

    if len(extract_numeric_cols(df=df))>1:

        correlation = calculate_correlation(df=df, method=correlation_method)

        plot = px.imshow(img=correlation,width=width, height=height,
                         title='Correlation plot using {} method'.format(correlation_method))

        figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_histogram(df,
                     numeric_cols,
                     nbins=20,
                     marginal=None,
                     width=None,
                     height=None,
                     cumulative=False,
                     histfunc='sum',
                     histnorm=None,
                     filename='histogram',
                     create_html=True,
                     orientation='v',
                     color=None,
                     opacity=1.0,
                     barmode='relative',
                     barnorm=None
                     ):
    """

    :param df: pandas dataframe
    :param numeric_cols:list of columns with numeric columns
    :param nbins:int, number of bins
    :param marginal:str, marginal must be one of 'rug','box','violin','histogram
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param cumulative:bool
    :param histfunc:str, histfunc must be one of 'count', 'sum', 'avg','min','max'
    :param histnorm:str, normalization method for histogram 'percent','probability','density','probability density'
    :param filename: str, name of file, the extension must be excluded
    :param create_html: boolean, whether to create an html file or not
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param color: str, column name to color the chart
    :param opacity: float Value between 0 and 1. Sets the opacity for markers
    :param barmode:
    :param barnorm:
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(numeric_cols, list), "numeric_cols must be a list"
    assert isinstance(nbins, int), "nbins must be an integer"
    assert marginal in ['rug','box','violin','histogram',None], "marginal options 'rug','box','violin','histogram'"
    assert isinstance(cumulative, bool), "cumulative must be a boolean"
    assert histfunc in ['count', 'sum', 'avg', 'min',
                        'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
    assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
    assert isinstance(filename, str), "filename must be a string without a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert orientation in ['v','h']
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None
    assert isinstance(opacity,float)
    assert 1.0 >= opacity >= 0.0
    assert barmode in ['group','overlay','relative']
    assert barnorm in [None,'fraction','percent']

    figure_list = list()

    if check_numeric(df=df,column_list=numeric_cols):
        for col in numeric_cols:
            plot = px.histogram(data_frame=df,
                                nbins=nbins,
                                x=col,
                                width=width,
                                height=height,
                                marginal=marginal,
                                cumulative=cumulative,
                                histfunc=histfunc,
                                histnorm=histnorm,
                                orientation=orientation,
                                title='Distribution of {}'.format(col),
                                color=color,
                                opacity=opacity,
                                barmode=barmode,
                                barnorm=barnorm
                                )
            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_scatter(df,
                   col_types,
                   filename='scatter',
                   marginal_x=None,
                   width=None,
                   height=None,
                   marginal_y=None,
                   log_x=False,
                   log_y=False,
                   orientation='v',
                   size=None,
                   opacity=1.0,
                   maximum_color_groups=5,
                   create_html=True):
    """
    :param df:pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param filename:str, name of file, the extension must be excluded
    :param marginal_x:options for including marginal charts on x axis
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param marginal_y:options for including marginal charts on y axis
    :param log_x:boolean, whether to create a log axis
    :param log_y:boolean, whether to create a log axis
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param create_html: boolean, whether to create an html file or not
    :param opacity: float, value between 0 and 1. Sets the opacity for markers
    :param maximum_color_groups: int, maximum number of color groups in the scatter plot
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(log_y, bool)
    assert isinstance(log_x, bool)
    assert marginal_x in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert marginal_y in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert isinstance(opacity, float), "opacity must be a float"
    assert opacity <= 1.0, "opacity must be between 0 and 1"
    assert opacity >= 0.0,"opacity must be a float between 0 and 1"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert orientation in ['v', 'h']
    assert isinstance(maximum_color_groups, int)
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()
    numeric_cols = col_types['numeric']

    if len(numeric_cols) < 2:
        # scatter plots need at least 2 columns of numeric type
        return figure_list
    plot_pairs = create_pairs(df, numeric_cols=numeric_cols)

    columns_with_distinct = columns_with_distinct_values(df,
                                                         maximum_number_distinct_values=maximum_color_groups,
                                                         categorical_only=True)
    if len(columns_with_distinct) == 0:
        columns_with_distinct.append(None)

    for pair in plot_pairs:
        x,y = pair
        for color in columns_with_distinct:
            plot = px.scatter(data_frame=df,
                              x=x,
                              y=y,width=width, height=height,
                              log_y=log_y,
                              log_x=log_x,
                              opacity=opacity,
                              orientation=orientation,
                              marginal_x=marginal_x,
                              marginal_y=marginal_y,
                              size=size,
                              color=color,
                              title='Plot of {} vs {} with color {}'.format(x,y,color))
            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_density_chart(df,
                         col_types,
                         nbinsx=None,
                         nbinsy=None,
                         width=None,
                         height=None,
                         maximum_color_groups=4,
                         typr_of_chart='contour',
                         filename='density_charts',
                         orientation='v',
                         marginal_x=None,
                         marginal_y=None,
                         log_x=False,
                         log_y=False,
                         histfunc='sum',
                         histnorm=None,
                         create_html=True
                         ):
    """
    Create density chart and density heatmap charts
    :param df: pandas dataframe
    :param col_types:  dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param nbinsx: int, number of bins on x axis
    :param nbinsy: int, number of bins on y axis
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :param maximum_color_groups: int, maximum number of color groups in the density charts
    :param typr_of_chart: str, 'contour' or 'heatmap'
    :param filename: str, name of file, the extension must be excluded
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param marginal_x: options for including marginal charts on x axis, one of 'rug', 'box', 'violin', 'histogram',None
    :param marginal_y: options for including marginal charts on y axis, one of 'rug', 'box', 'violin', 'histogram',None
    :param log_x: boolean, whether to create a log x axis
    :param log_y: boolean, whether to create a log y axis
    :param histfunc: histfunc must be one of 'count', 'sum', 'avg','min','max'
    :param histnorm: str, normalization method for histogram 'percent','probability','density','probability density'
    :param create_html: boolean, whether to create an html file or not
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(nbinsx,int) or nbinsx is None
    assert isinstance(nbinsy, int) or nbinsy is None
    assert isinstance(log_y, bool)
    assert isinstance(log_x, bool)
    assert marginal_x in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert marginal_y in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert histfunc in ['count', 'sum', 'avg', 'min',
                        'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
    assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
    assert orientation in ['v', 'h']
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert typr_of_chart in ['contour','heatmap']
    assert isinstance(width, int) or width is None
    assert isinstance(height, int) or height is None

    figure_list = list()
    numeric_cols = col_types['numeric']

    if len(numeric_cols) < 2:
        # density contour at least 2 columns of numeric type
        return figure_list

    # create pairing for x and y values
    plot_pairs = create_pairs(df, numeric_cols=numeric_cols)

    columns_with_distinct = columns_with_distinct_values(df,
                                                         maximum_number_distinct_values=maximum_color_groups,
                                                         categorical_only=True)

    if len(columns_with_distinct) == 0:
        columns_with_distinct.append(None)

    for pair in plot_pairs:
        x,y = pair
        for z in numeric_cols:
            if typr_of_chart == 'contour':
                for color in columns_with_distinct:
                    title = "Density contour of {} vs {},with {} on z axis".format(x, y, z)
                    plot = px.density_contour(data_frame=df,
                                              x=x,
                                              y=y,
                                              z=z,width=width, height=height,
                                              color=color,
                                              histnorm=histnorm,
                                              histfunc=histfunc,
                                              nbinsx=nbinsx,
                                              nbinsy=nbinsy,
                                              log_y=log_y,
                                              log_x=log_x,
                                              marginal_x=marginal_x,
                                              marginal_y=marginal_y,
                                              orientation=orientation,
                                              title=title)
                    figure_list.append(plot)

            if typr_of_chart == 'heatmap':
                title = "Density heatmap of {} vs {},with {} on z axis".format(x, y, z)

                plot = px.density_heatmap(data_frame=df,
                                          x=x,
                                          y=y,
                                          z=z,
                                          width=width,
                                          height=height,
                                          histnorm=histnorm,
                                          histfunc=histfunc,
                                          nbinsx=nbinsx,
                                          nbinsy=nbinsy,
                                          log_y=log_y,
                                          log_x=log_x,
                                          marginal_x=marginal_x,
                                          marginal_y=marginal_y,
                                          orientation=orientation,
                                          title=title)
                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_bar(df,
               col_types,
               maximum_number_bars=10,
               color=None,
               orientation='v',
               opacity=1.0,
               barmode='group',
               filename='barcharts',
               width=None,
               height=None,
               log_x=False,
               log_y=False,
               create_html=True):
    """
    Creates barcharts for numeric data as they compare across categories
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param maximum_number_bars: int, maximum number of bars in all charts
    :param color: str, column name to color the chart
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param opacity:
    :param barmode:
    :param filename::str, filename for the html file
    :param width: int, width of chart in pixels
    :param height: int, height of chart in pixels
    :return: list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(maximum_number_bars,int), "maximum_number_bars must be an int"
    combined = list()
    categorical = col_types['categorical']
    object_cols = col_types['object']
    boolean_cols = col_types['boolean']
    combined.extend(categorical)
    combined.extend(object_cols)
    combined.extend(boolean_cols)
    combined.extend([None])
    assert color in combined,"color column must be categorical in nature"
    assert orientation in ['v', 'h'], "orientation must be a string with either v or h"
    assert opacity <= 1.0, "opacity must be a number between 0 and 1"
    assert opacity >= 0.0, "opacity must be a number between 0 and 1"
    assert barmode in ['group', 'overlay', 'relative']
    assert isinstance(filename, str), "filename must be a string without a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(width, int) or width is None,"width must be an int or None"
    assert isinstance(height, int) or height is None,"height must be an int or None"
    assert isinstance(log_y,bool)
    assert isinstance(log_x,bool)
    assert isinstance(create_html,bool)

    figure_list = list()
    numeric_cols = col_types['numeric']

    categorical_columns = columns_with_distinct_values(df=df,
                                                       maximum_number_distinct_values=maximum_number_bars)

    if len(numeric_cols) != 0 and len(categorical_columns) != 0:
        for col in numeric_cols:
            for category in categorical_columns:

                plot = px.bar(data_frame=df,
                              x=category,
                              y=col,
                              color=color,
                              opacity=opacity,
                              orientation=orientation,
                              barmode=barmode,
                              log_y=log_y,
                              log_x=log_x,
                              width=width,
                              height=height,
                              title="Barchart of {} across {} categories".format(col,category)
                              )

                figure_list.append(plot)

    if create_html:

        figures_to_html(figs=figure_list, filename=filename)

    return figure_list






















import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from .utils import (check_columns,
                    check_dataframe,)


class ColumnDropper(BaseEstimator, TransformerMixin):
    def __init__(self,
                 column_list):
        """
        This class drops columns from a dataframe, this operation is done inplace.

        :param column_list: list which contains the names of columns in the dataframe to be dropped

        Usage
        >>> from datamallet.tabular.preprocess import ColumnDropper
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],
                   'B':[2,4,6,8,10],
                   'C':['dog','cat', 'sheep','dog','cat'],
                   'D':['male','male','male','female','female'],
                   'E':[True,True,False,True,True]})

        # initialize the ColumnDropper class
        >>> column_dropper = ColumnDropper(column_list=['C','D','E'])

        # call the transform method
        >>> dropped_df = column_dropper.transform(X=df)

        # check
        >>> print(dropped_df.columns)
        Index(['A', 'B'], dtype='object')

        """
        self.column_list = column_list

    def fit(self, X, y=None):
        assert (isinstance(X, pd.DataFrame)), 'X needs to be a pandas dataframe'
        return self

    def transform(self, X, y=None):
        if check_columns(X, self.column_list) and check_dataframe(X):
            X = X.copy()
            X.drop(labels=self.column_list, inplace=True, axis=1)
        return X


class ColumnRename(BaseEstimator, TransformerMixin):
    def __init__(self, rename_dictionary):
        """
        Rename columns in place
        :param rename_dictionary: python dictionary with the old names as keys and new names as value.

        Usage
        >>> from datamallet.tabular.preprocess import ColumnRename
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

        >>> rename_dictionary={'A':'V','B':'W'}

        >>> columnrenamer = ColumnRename(rename_dictionary=rename_dictionary)
        >>> renamed_df = columnrenamer.transform(X=df)

        >>> print(renamed_df.columns)
        Index(['V', 'W', 'C', 'D', 'E'], dtype='object')

        """
        self.rename_dictionary = rename_dictionary

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        assert isinstance(self.rename_dictionary, dict), 'Rename dictionary must be a dictionary'

        col_list = list()
        for old_name, new_name in self.rename_dictionary.items():
            col_list.append(old_name)

        if check_columns(df=X, column_list=col_list) and check_dataframe(df=X):
            X = X.copy()
            X.rename(columns=self.rename_dictionary, inplace=True)

        return X


class FunctionMapper(BaseEstimator, TransformerMixin):
    def __init__(self,
                 func,
                 axis='columns',
                 new_col_name=None,
                 col_name=None):
        """
        Rename columns in place
        :param func: python function to be applied on dataframe.
        :param axis: str, axis to apply the function.
        :param new_col_name: str, column name of new column.
        :param col_name: str, column name to apply.

        Usage
        >>> from datamallet.tabular.preprocess import FunctionMapper
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})


        """
        self.func = func
        self.axis = axis
        self.col_name = col_name
        self.new_col_name = new_col_name

    def fit(self, X, y=None):
        assert self.axis in ['columns', 'index']
        return self

    def transform(self, X, y=None):
        if check_dataframe(df=X):
            X = X.copy()
            if self.col_name is None:
                X[self.new_col_name] = X.apply(self.func, axis=self.axis)
            if check_columns(df=X,column_list=[self.col_name]):
                X[self.new_col_name] = X[self.col_name].apply(self.func)

        return X


class ColumnSelector(BaseEstimator, TransformerMixin):
    def __init__(self,
                 column_list
                 ):
        """
        Selects certain columns in a dataframe
        :param column_list: list of column names.

        Usage
        >>> from datamallet.tabular.preprocess import ColumnSelector
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})


        """
        self.column_list = column_list

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_dataframe(df=X):
            X = X.copy()

            if check_columns(df=X,column_list=self.column_list):
                X= X.loc[:,self.column_list]

        return X





